/* Copyright (C) 2008 DDengine.org

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#ifdef USE_PRAGMA_IMPLEMENTATION
#pragma implementation        // gcc: Class implementation
#endif

#define MYSQL_SERVER 1
#include "mysql_priv.h"
#include <mysql/plugin.h>
#include "double_table_handler.h"
#include "single_table_handler.h"
#include "ha_revision.h"

const char revision_suffix[] = "_revision";

extern handlerton *revision_hton;

double_table_handler::double_table_handler(TABLE_SHARE *table_arg,
                                           MEM_ROOT *root)
  : table(NULL), uhandler(NULL), rev_handler(NULL),
    table_handler(NULL),
    select_mode(revision_handler::none),
    mem_root(root), opened(FALSE)
{
  if (table_arg && table_arg->db_type()) 
  {
    create(table_arg);
  }
}


void double_table_handler::set_select_mode(select_mode_t mode)
{
  select_mode= mode;
}


void double_table_handler::set_timestamp(const MYSQL_TIME &t)
{
  long tz;
  my_bool in_dst;
  timestamp= my_system_gmt_sec(&t, &tz, &in_dst);
}


handlerton *get_underlying_handlerton(THD *thd, const LEX_STRING &comment)
{
  handlerton *underlying= 0;

  if (comment.length != 0)
  {
    const char *sep= strstr(comment.str, separator);
    if (sep != NULL)
    {
      LEX_STRING name;
      name.str= comment.str;
      name.length= sep - comment.str;

      plugin_ref plugin= ha_resolve_by_name(thd, &name);
      if (plugin)
        underlying= plugin_data(plugin, handlerton *);
      
      if (underlying == NULL)
      {
        char warn_buff[MYSQL_ERRMSG_SIZE];
        my_snprintf(warn_buff, sizeof(warn_buff),
                    ER(ER_UNKNOWN_STORAGE_ENGINE),
                    comment.str);
        push_warning(current_thd, MYSQL_ERROR::WARN_LEVEL_WARN,
                     ER_UNKNOWN_STORAGE_ENGINE, warn_buff);
      }
    }
  }
  if (underlying == NULL)
  {
    underlying= ha_default_handlerton(thd);
  }
  return underlying;
}


/** @brief
    Intantiates the underlying handler based on the table comment field
    if found. Otherwise use the default handler.
*/
void double_table_handler::create(TABLE_SHARE *table_arg)
{
  THD *thd= current_thd;
  handlerton *underlying= get_underlying_handlerton(thd, table_arg->comment);
  uhandler= underlying->create(underlying,
                               table_arg,
                               mem_root);
}


double_table_handler::~double_table_handler()
{
  delete uhandler;
}


int double_table_handler::open_table()
{
  TABLE_LIST table_list;
  enum thr_lock_type lock_type=
    reinterpret_cast<TABLE_LIST *>
    (table->in_use->lex->select_lex.table_list.first)->lock_type;
  LEX_STRING rev_name;
  char path[FN_REFLEN];
  path[0]= 0;
  rev_name.str= path;
  rev_name.length= (strxmov(path, table->s->table_name.str,
                            revision_suffix, NullS) - path);
  table_list.init_one_table(table->s->db.str, path, lock_type);
  bool refresh= FALSE;
  uint flags= 0;

  TABLE *t= ::open_table(current_thd, &table_list, mem_root, &refresh, flags);

  if (t == NULL)
    return (HA_ERR_NO_SUCH_TABLE);

  rev_handler= static_cast<ha_revision *>(t->file);
  table_handler=
    static_cast<single_table_handler *> (rev_handler->get_uhandler());

  opened= TRUE;
  return (0);
}


int double_table_handler::reset()
{
  int r= uhandler->ha_reset();
  opened= FALSE;
  return (r);
}


int double_table_handler::open(TABLE *tab, const char *name, int mode,
                               uint test_if_locked)
{
  DBUG_ENTER("double_table_handler::open");

  table= tab;
  int r= uhandler->ha_open(tab, name, mode, test_if_locked);
  DBUG_RETURN(r);
}


int double_table_handler::close()
{
  DBUG_ENTER("double_table_handler::close");
  DBUG_RETURN(uhandler->close());
}


int double_table_handler::index_read(uchar * buf, const uchar * key,
                                     uint key_len,
                                     enum ha_rkey_function find_flag)
{
  DBUG_ENTER("double_table_handler::index_read");
  key_part_map keypart_map= 0;
  for (uint i= 0, len= 0; len < key_len;
       len+= table->key_info[uhandler->get_index()].key_part[i].store_length,
         ++i)
  {
    keypart_map+= (1 << i);
  }
  DBUG_RETURN(uhandler->index_read_map(buf, key,
                                       keypart_map,
                                       find_flag));
}


int double_table_handler::index_next(uchar *buf)
{
  DBUG_ENTER("double_table_handler::index_next");
  int i= uhandler->index_next(buf);
  DBUG_RETURN(i);
}


int double_table_handler::rnd_next(uchar *buf)
{
  DBUG_ENTER("double_table_handler::rnd_next");
  int i= uhandler->rnd_next(buf);
  DBUG_RETURN(i);
}


double double_table_handler::read_time(uint index, uint ranges, ha_rows rows)
{
  return uhandler->read_time(index, ranges, rows);
}


double double_table_handler::scan_time()
{
  return uhandler->scan_time();
}


uint double_table_handler::lock_count(void) const
{
  return 2;
}


THR_LOCK_DATA **double_table_handler::store_lock(THD *thd,
                                                 THR_LOCK_DATA **to,
                                                 enum thr_lock_type lock_type)
{
  if (!opened)
  {
    int r= open_table();
    if (r != 0)
      return 0;
  }
  to= uhandler->store_lock(thd, to, lock_type);
  return table_handler->store_lock(thd, to, lock_type);
}


int double_table_handler::external_lock(THD *thd, int lock_type)
{
  return uhandler->ha_external_lock(thd, lock_type);
}


bool double_table_handler::get_error_message(int, String *)
{
  return TRUE;
}


int double_table_handler::write_row(uchar * buf)
{
  DBUG_ENTER("double_table_handler::write_row");
  int r= uhandler->ha_write_row(buf);
  DBUG_RETURN(r);
}


int double_table_handler::update_row(const uchar *old_data,
                                     uchar * new_data)
{
  DBUG_ENTER("double_table_handler::update_row");

  if (!opened)
  {
    int r= open_table();
    if (r != 0)
      DBUG_RETURN(r);
  }

  int r= table_handler->write_or_update_row(table->field, old_data, new_data);
  if (r != 0)
    DBUG_RETURN(r);
  r= uhandler->ha_update_row(old_data, new_data);
  DBUG_RETURN(r);
}


int double_table_handler::delete_row(const uchar *data)
{
  DBUG_ENTER("double_table_handler::delete_row");

  if (!opened)
  {
    int r= open_table();
    if (r != 0)
      DBUG_RETURN(r);
  }

  int r= table_handler->delete_row(table->field, data);

  if (r != 0)
    DBUG_RETURN(r);
  r= uhandler->ha_delete_row(data);

  DBUG_RETURN(r);
}


void double_table_handler::delete_table(const char *name, TABLE_SHARE *table)
{

  DBUG_ENTER("double_table_handler::delete_table");

  create(table);
  uhandler->ha_delete_table(name);

  {
    /* Delete revision table as well.
    */
    const uint size = 32;
    char type[size];
    LEX_STRING nm = { type, size };
    const char *sep= strstr(table->comment.str, separator);
    if (sep != NULL)
    {
      nm.length= sep - table->comment.str;
      strncpy(type, table->comment.str, nm.length);
      type[nm.length]= 0;
    }
    else
    {
      nm.str= table->comment.str;
      nm.length= table->comment.length;
    }

    THD *thd= current_thd;

    const char *tname= strstr(name, thd->db) + 1 + strlen(thd->db);

    char path[FN_REFLEN];
    path[0]= 0;
    strxmov(path, tname, revision_suffix, NullS);

    mysql_ha_flush(thd);

    bool b= remove_table_from_cache(thd, thd->db, path, 0);
    if (b)
    {
      push_warning(thd, MYSQL_ERROR::WARN_LEVEL_WARN,
                   ER_BAD_TABLE_ERROR, "");
      return;
    }

    handlerton *underlying=
      plugin_data(ha_resolve_by_name(thd, &nm), handlerton *);
    if (underlying == 0)
    {
      char warn_buff[MYSQL_ERRMSG_SIZE];
      my_snprintf(warn_buff, sizeof(warn_buff), ER(ER_UNKNOWN_STORAGE_ENGINE),
                  nm.str);
      push_warning(thd, MYSQL_ERROR::WARN_LEVEL_WARN,
                   ER_UNKNOWN_STORAGE_ENGINE, warn_buff);
    }
    else
    {
      TABLE_SHARE tab;
      bzero(static_cast<void *>(&tab), sizeof(tab));

      char rev_name[FN_REFLEN];
      rev_name[0]= 0;
      strxmov(rev_name, name, revision_suffix, NullS);

      /* Delete REV & frm files as well
      */
      ha_revision ha(revision_hton, &tab, mem_root);
      int r= ha.delete_table(rev_name);
      if (r != 0)
      {
        push_warning(thd, MYSQL_ERROR::WARN_LEVEL_ERROR, r, "");
      }
      strxmov(rev_name, rev_name, ".frm", NullS);
      my_delete(rev_name,MYF(MY_WME));
    }
  }
  DBUG_VOID_RETURN;
}


int double_table_handler::create(const char *name, TABLE *table_arg,
                                 HA_CREATE_INFO *create_info)
{

  DBUG_ENTER("double_table_handler::create");  

  /* Comment string contains the underlying storage engine.
  */
  if (uhandler == NULL)
  {
    create(table_arg->s);
  }
  uhandler->init();

  int r= uhandler->ha_create(name, table_arg, create_info);

  if (r != 0)
  {
    DBUG_RETURN(r);
  }

  /* Create the table containing the versioned data.
  */

  /* Table name
  */
  LEX_STRING rev;
  char rev_name[FN_REFLEN];
  rev_name[0]= 0;
  rev.str= rev_name;
  rev.length= (strxmov(rev_name, name, revision_suffix, NullS)
                    - rev_name);

  const LEX_STRING old_path= table_arg->s->path;

  table_arg->s->path= rev;

  /* Table normalized path
  */
  LEX_STRING rev_npath;
  char rev_name_npath[FN_REFLEN];
  rev_name_npath[0]= 0;
  rev_npath.str= rev_name_npath;
  rev_npath.length= (strxmov(rev_name_npath,
                             table_arg->s->normalized_path.str,
                             revision_suffix, NullS)
                     - rev_name_npath);

  const LEX_STRING old_npath= table_arg->s->normalized_path;
  table_arg->s->normalized_path= rev_npath;

  /* Revision table is a single table
  */
  LEX_STRING rcomment;
  char rev_comment[FN_REFLEN];
  rev_comment[0]= 0;
  const char *sep= strstr(table_arg->s->comment.str, separator);
  rcomment.length= sep - table_arg->s->comment.str;
  strncpy(rev_comment, table_arg->s->comment.str, rcomment.length);
  rev_comment[rcomment.length]= 0;
  rcomment.str= rev_comment;

  const LEX_STRING old_rcomment= table_arg->s->comment;
  table_arg->s->comment= rcomment;

  const LEX_STRING crcomment= create_info->comment;
  create_info->comment= rcomment;

  ha_revision ha(revision_hton, table_arg->s, mem_root);

  r= ha.create(rev_name, table_arg, create_info);

  table_arg->s->path= old_path;
  table_arg->s->comment= old_rcomment;
  create_info->comment= crcomment;
  table_arg->s->normalized_path= old_npath;

  DBUG_RETURN(r);
}
