/* Copyright (C) 2008 DDengine.org

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#ifdef USE_PRAGMA_IMPLEMENTATION
#pragma implementation        // gcc: Class implementation
#endif

#define MYSQL_SERVER 1
#include "mysql_priv.h"
#include <mysql/plugin.h>
#include "single_table_handler.h"

const char revision_id[]= "revision_id";
const char revision_ts[]= "revision_timestamp";
const char revision_del[]= "revision_deleted";
const uint deleted_flag= 1;
const uint revision_fields_count= 3;

/** @brief This class is used when looking for specific rows depending
   on the select mode set in the system variable.
 */
class select_handler
{

public:
  virtual ~select_handler() { }

  /** Returns true if the row matches and false otherwise.
  */
  virtual int match(uchar *buf) = 0;
  
  /** Read the index and returns the corresponding row.
     The index is the primary key.
     The only missing part of the key is the revisoin_id.
     The key_map does not contain the revision_id.
  */
  virtual int index_read_with_pk(handler *uhandler, uchar *buf,
                                 const uchar *key,
                                 KEY *key_info,
                                 const key_part_map &key_map) = 0;

  /** Returns true if the handler looks for a specific revision,
      which will use index_read_with_pk() and false otherwise.
   */
  virtual bool look_for_specific_revision() { return(TRUE); }

};

namespace
{

  /** Returns true if the revision is deleted and false otherwise.
  */
  bool row_is_deleted(TABLE *table, uchar *record,
                      Field *revision_del_field)
  {
    my_ptrdiff_t offset= revision_del_field->ptr - table->record[0];
    Field_tiny deleted_field
      (record + offset,
       revision_del_field->field_length,
       revision_del_field->null_ptr,
       revision_del_field->null_bit,
       revision_del_field->unireg_check,
       revision_del_field->field_name,
       static_cast<Field_tiny *>(revision_del_field)->zerofill,
       static_cast<Field_tiny *>(revision_del_field)->unsigned_flag);
    deleted_field.table= revision_del_field->table;
 
    my_bitmap_map *old_map= tmp_use_all_columns(table, table->read_set);
    const uint d= deleted_field.val_int();
    tmp_restore_column_map(table->read_set, old_map);
    if (deleted_flag == d)
      return TRUE;
    return FALSE;
  }


  /** Returns the timestamp revision of the row.
  */
  my_time_t revision_timestamp(TABLE *table, uchar *record,
                               Field *revision_ts_field)
  {
    my_ptrdiff_t offset= revision_ts_field->ptr - table->record[0];
    Field_timestamp ts_field
      (record + offset,
       revision_ts_field->field_length,
       revision_ts_field->null_ptr,
       revision_ts_field->null_bit,
       revision_ts_field->unireg_check,
       revision_ts_field->field_name,
       table->s,
       table->s->table_charset);
    ts_field.table= revision_ts_field->table;

    my_bitmap_map *old_map= tmp_use_all_columns(table, table->read_set);
    my_bool null= 0;
    const my_time_t t= ts_field.get_timestamp(&null);
    tmp_restore_column_map(table->read_set, old_map);
    return t;
  }

  /** @brief This class is used when no specific row are to be returned:
     all rows match.
     This handler is used in UPDATE and DELETE statements.
  */
  class none_select_handler : public select_handler
  {

  public:
    virtual int match(uchar *buf);
    virtual int index_read_with_pk(handler *uhandler,
                                   uchar *buf, const uchar *key,
                                   KEY *key_info,
                                   const key_part_map &key_map);
  };

  /** @brief This class is used when only the current row is to be returned
     (last revision): non deleted rows match.
  */
  class current_select_handler : public select_handler
  {

  public:
    current_select_handler(TABLE *tab, Field *del_fld)
      : table(tab),
        del_field(del_fld)
      {
      }

    virtual int match(uchar *buf);
    virtual int index_read_with_pk(handler *uhandler,
                                   uchar *buf, const uchar *key,
                                   KEY *key_info,
                                   const key_part_map &key_map);

  private:
    TABLE *table;
    Field *del_field;
  };

  /** @brief This class is used when a row with a specific timestamp is to
     be returned.
  */
  class timestamp_select_handler : public select_handler
  {

  public:
    timestamp_select_handler(TABLE *tab, my_time_t ts,
                             Field_timestamp *ts_fld,
                             Field *del_fld, uchar *rec0, uchar *rec1)
      : timestamp(ts),
        table(tab),
        ts_field(ts_fld),
        del_field(del_fld),
        record0(rec0), record1(rec1)
      {
      }

    virtual int match(uchar *buf);
    virtual int index_read_with_pk(handler *uhandler,
                                   uchar *buf, const uchar *key,
                                   KEY *key_info,
                                   const key_part_map &key_map);

  private:
    my_time_t timestamp;
    TABLE *table;
    Field_timestamp *ts_field;
    Field *del_field;
    uchar *record0;
    uchar *record1;

  };

  /** @brief This class is used when deleted revisions are to be returned.
  */
  class deleted_select_handler : public select_handler
  {

  public:
    deleted_select_handler(TABLE *tab, Field *del_fld,
                           uchar *rec)
      : table(tab),
        del_field(del_fld),
        record(rec)
      {
      }

    virtual int match(uchar *buf);
    virtual int index_read_with_pk(handler *uhandler,
                                   uchar *buf, const uchar *key,
                                   KEY *key_info,
                                   const key_part_map &key_map);

  private:
    TABLE *table;
    Field *del_field;
    uchar *record;
  };

  /** @brief This class is used when no specific row are to be returned:
      all rows match and no index read with pk is performed.
  */
  class history_select_handler : public select_handler
  {

  public:
    virtual int match(uchar *buf);
    virtual int index_read_with_pk(handler *uhandler,
                                   uchar *buf, const uchar *key,
                                   KEY *key_info,
                                   const key_part_map &key_map);
    virtual bool look_for_specific_revision();
  };

}


int none_select_handler::match(uchar *)
{
  return(TRUE);
}


int none_select_handler::index_read_with_pk(handler *uhandler,
                                            uchar *buf, const uchar *key,
                                            KEY *,
                                            const key_part_map &keypart_map)
{
  return(uhandler->index_read_map(buf, key, keypart_map,
                              HA_READ_PREFIX_LAST));
}


int current_select_handler::match(uchar *buf)
{
  return(!row_is_deleted(table, buf, del_field));
}


int current_select_handler::index_read_with_pk(handler *uhandler,
                                               uchar *buf, const uchar *key,
                                               KEY *,
                                               const key_part_map &keypart_map)
{
  int i= uhandler->index_read_map(buf, key, keypart_map,
                                  HA_READ_PREFIX_LAST);
  if (i != 0)
    return(i);

  if (row_is_deleted(table, buf, del_field))
    return(HA_ERR_END_OF_FILE);

  return(0);
}


int timestamp_select_handler::match(uchar *buf)
{
  return((revision_timestamp(table, buf, ts_field) <= timestamp)
         && (!row_is_deleted(table, buf, del_field)));
}


int timestamp_select_handler::index_read_with_pk(handler *uhandler,
                                                 uchar *buf, const uchar *key,
                                                 KEY *key_info,
                                                 const key_part_map &key_map)
{
  if (revision_timestamp(table, buf, ts_field) > timestamp)
    return(HA_ERR_END_OF_FILE);

  memcpy(record0, buf, table->s->reclength);

  /* Store previous key to go backward to last matching row
   */
  KEY full_key_info(*key_info);

  key_part_map full_keypart_map=
    make_prev_keypart_map(full_key_info.key_parts);

  key_copy(record1, buf, &full_key_info, 0);

  /* Compare without the revision id.
  */
  --key_info->key_parts;

  /* Look for the last row which satisfies:
     row_timestamp <= timestamp and next_row_timestamp > timestamp.
     If the row is deleted, no record is found.
  */
  int i= 0;
  while (!key_rec_cmp(key_info, buf, record0)
         && (revision_timestamp(table, buf, ts_field) <= timestamp))
  {
    key_copy(record1, buf, &full_key_info, 0);
    i= uhandler->index_next(buf);
    if (i == HA_ERR_END_OF_FILE)
    {
      i= uhandler->index_read_map(buf, record1, full_keypart_map,
                                  HA_READ_KEY_EXACT);

      if ((i != 0)
          || (row_is_deleted(table, buf, del_field)))
        return(HA_ERR_END_OF_FILE);

      (void)uhandler->index_read_map(record1, key, key_map,
                                     HA_READ_PREFIX_LAST);
      return(i);
    }
  }
  i= uhandler->index_read_map(buf, record1, full_keypart_map,
                              HA_READ_KEY_EXACT);
  (void)uhandler->index_read_map(record1, key, key_map,
                                 HA_READ_PREFIX_LAST);
  if (i != HA_ERR_END_OF_FILE)
  {
    if ((key_rec_cmp(key_info, buf, record0))
        || (row_is_deleted(table, buf, del_field)))
      return(HA_ERR_END_OF_FILE);
  }
  return(i);
}


int deleted_select_handler::match(uchar *buf)
{
  return(row_is_deleted(table, buf, del_field));
}


int deleted_select_handler::index_read_with_pk(handler *uhandler,
                                               uchar *buf, const uchar *,
                                               KEY *key_info,
                                               const key_part_map &)
{

  /* Compare without revision_id
  */
  --key_info->key_parts;

  memcpy(record, buf, table->s->reclength);
  int i= 0;

  /* check if deleted flag is set
  */
  while (!key_rec_cmp(key_info, buf, record)
         && (!row_is_deleted(table, buf, del_field)))
  {
    i= uhandler->index_next(buf);
    if (i == HA_ERR_END_OF_FILE)
      return(i);
  }
  if (key_rec_cmp(key_info, buf, record))
    return(HA_ERR_END_OF_FILE);
  return(i);
}


int history_select_handler::match(uchar *)
{
  return TRUE;
}


int history_select_handler::index_read_with_pk(handler *,
                                               uchar *, const uchar *,
                                               KEY *,
                                               const key_part_map &)
{
  return(0);
}


bool history_select_handler::look_for_specific_revision()
{
  return(FALSE);
}


single_table_handler::single_table_handler(TABLE_SHARE *table_arg,
                                           MEM_ROOT *root)
  : table(NULL), uhandler(NULL), revision_field(NULL),
    revision_ts_field(NULL), revision_del_field(NULL),
    select_mode(revision_handler::none),
    mem_root (root), shandler(new none_select_handler),
    no_field_check(FALSE)
{
  record[0]= record[1]= NULL;

  if (table_arg && table_arg->db_type()) 
  {
    create(table_arg);
  }
}


void single_table_handler::set_select_mode(select_mode_t mode)
{

  if ((select_mode == mode)
      && (select_mode != revision_handler::timestamp))
    return;

  select_mode= mode;

  delete shandler;

  switch (select_mode)
  {
  case revision_handler::current:
    shandler= new current_select_handler(table,
                                         revision_del_field);
    break;
  case revision_handler::deleted:
    shandler= new deleted_select_handler(table,
                                         revision_del_field,
                                         record[0]);
    break;
  case revision_handler::timestamp:
    shandler= new timestamp_select_handler(table,
                                           timestamp, revision_ts_field,
                                           revision_del_field,
                                           record[0], record[1]);
    break;
  case revision_handler::history:
    shandler= new history_select_handler;
    break;
  case revision_handler::none:
  default:
    shandler= new none_select_handler;
  }
}


void single_table_handler::set_timestamp(const MYSQL_TIME &t)
{
  long tz;
  my_bool in_dst;
  timestamp= my_system_gmt_sec(&t, &tz, &in_dst);
}

/** @brief
    Intantiates the underlying handler based on the table comment field
    if found. Otherwise use the default handler.
*/
void single_table_handler::create(TABLE_SHARE *table_arg)
{
  handlerton *underlying= 0;
  THD *thd= current_thd;
  if (table_arg->comment.length != 0)
  {
    plugin_ref plugin=
      ha_resolve_by_name(thd, &table_arg->comment);
    if (plugin)
      underlying= plugin_data(plugin, handlerton *);

    if (underlying == NULL)
    {
      char warn_buff[MYSQL_ERRMSG_SIZE];
      my_snprintf(warn_buff, sizeof(warn_buff), ER(ER_UNKNOWN_STORAGE_ENGINE),
                  table_arg->comment.str);
      push_warning(current_thd, MYSQL_ERROR::WARN_LEVEL_WARN,
                   ER_UNKNOWN_STORAGE_ENGINE, warn_buff);
    }
  }
  if (underlying == NULL)
  {
    underlying= ha_default_handlerton(thd);
  }
  uhandler= underlying->create(underlying,
                               table_arg,
                               mem_root);
}


single_table_handler::~single_table_handler()
{
  if ((NULL != record[0])
      && (NULL != record[1]))
  {
    if (record[0] < record[1])
      my_free(record[0], MYF(0));
    else
      my_free(record[1], MYF(0));
  }
  delete uhandler;
  delete shandler;
}


/** @brief
    Checks that there is no revision field set to be written to the table.
*/
bool single_table_handler::check_revision_field() const
{
  if (bitmap_is_set(table->write_set, revision_field->field_index))
    return TRUE;
  if (bitmap_is_set(table->write_set, revision_ts_field->field_index))
    return TRUE;
  if (bitmap_is_set(table->write_set, revision_del_field->field_index))
    return TRUE;
  return FALSE;
}


int single_table_handler::open(TABLE *tab, const char *name, int mode,
                               uint test_if_locked)
{
  DBUG_ENTER("single_table_handler::open");

  table= tab;

  if (alloc_records())
    DBUG_RETURN(HA_ERR_OUT_OF_MEM);

  int r= uhandler->ha_open(tab, name, mode, test_if_locked);

  uint fields= 0;

  for (Field **field= table->field; *field; field++)
  {
    const char *fn= (*field)->field_name;
    if (strcmp(fn, revision_id) == 0)
    {
      revision_field= *field;
      ++fields;
    }
    else if (strcmp(fn, revision_ts) == 0)
    {
      revision_ts_field= static_cast<Field_timestamp *>(*field);
      ++fields;
    }
    else if (strcmp(fn, revision_del) == 0)
    {
      revision_del_field= *field;
      ++fields;
    }
  }

  /* Missing field
  */
  if (fields != revision_fields_count)
  {
    char warn_buff[MYSQL_ERRMSG_SIZE];
    my_snprintf(warn_buff, sizeof(warn_buff),
                ER(ER_FIELD_NOT_FOUND_PART_ERROR));
    push_warning(current_thd, MYSQL_ERROR::WARN_LEVEL_WARN,
                 ER_FIELD_NOT_FOUND_PART_ERROR, warn_buff);
  }

  DBUG_RETURN(r);
}


int single_table_handler::reset()
{
  return (uhandler->ha_reset());
}


int single_table_handler::close()
{
  DBUG_ENTER("single_table_handler::close");
  DBUG_RETURN(uhandler->close());
}


enum {
  HA_ERR_REVISION_READ_ONLY_FIELD = 10000,
  HA_ERR_REVISION_DUPLICATE_FIELD,
  HA_ERR_REVISION_DELETED_ROW,
  HA_ERR_REVISION_MISSING_FIELD
};


int single_table_handler::index_read(uchar * buf, const uchar * key,
                                     uint key_len,
                                     enum ha_rkey_function find_flag)
{
  DBUG_ENTER("single_table_handler::index_read");

  const uint index= uhandler->get_index();

  if (shandler->look_for_specific_revision())
  {
    /* If wanted revision is current,
       Check if we are using a primary key:
       a revision can be uniquely determined from the pk,
       a unique index is not enough: two different rows with the same index
       value but different revisions can use two different pk values.
    */
    if (index == table->s->primary_key)
    {

      KEY key_info= table->key_info[index];

      /* If the only missing part is the revision,
         find the latest matching revision.
         Otherwise we are not able to match a particular revision of a row.
      */
      const uint rev_len= key_info.key_part[key_info.key_parts - 1].length;
      if (key_len == key_info.key_length - rev_len)
      {

        /* If revision_id is specified in the where clause, we do not
           look for the latest revision.
        */
        THD *thd= current_thd;
        COND *cond= thd->lex->select_lex.where;
        bool found= FALSE;

        while (cond != NULL)
        {
          if ((cond->name != NULL)
              && (strcmp(cond->name, revision_id) == 0))
          {
            found= TRUE;
            break;
          }
          cond= cond->next;
        }

        /* Revision id is not set, we look for the specific revision
        */
        if (!found)
        {

          const key_part_map keypart_map=
            make_prev_keypart_map(key_info.key_parts - 1);

          int j= uhandler->index_read_map(buf, key, keypart_map, find_flag);
          if (j == HA_ERR_END_OF_FILE)
            DBUG_RETURN(j);

          DBUG_RETURN(shandler->index_read_with_pk(uhandler, buf, key,
                                                   &key_info, keypart_map));
        }
      }
    }
  }

  key_part_map keypart_map= 0;
  for (uint i= 0, len= 0; len < key_len;
       len+= table->key_info[index].key_part[i].store_length, ++i)
  {
    keypart_map+= (1 << i);
  }

  int i= uhandler->index_read_map(buf, key, keypart_map, find_flag);

  if (i != 0)
    DBUG_RETURN(i);

  while (!shandler->match(buf))
  {
    i= uhandler->index_next(buf);
    if (i != 0)
      DBUG_RETURN(i);
  }
  DBUG_RETURN(i);
}


int single_table_handler::index_next(uchar *buf)
{
  DBUG_ENTER("single_table_handler::index_next");

  int i= uhandler->index_next(buf);
  if (i != 0)
    DBUG_RETURN(i);

  while (!shandler->match(buf))
  {
    i= uhandler->index_next(buf);
    if (i != 0)
      DBUG_RETURN(i);
  }
  DBUG_RETURN(i);
}


int single_table_handler::rnd_next(uchar *buf)
{
  DBUG_ENTER("single_table_handler::rnd_next");

  int i= uhandler->rnd_next(buf);

  if (i != 0)
    DBUG_RETURN(i);

  while (!shandler->match(buf))
  {
    i= uhandler->rnd_next(buf);
    if (i != 0)
      DBUG_RETURN(i);
  }
  DBUG_RETURN(i);
}


double single_table_handler::read_time(uint index, uint ranges, ha_rows rows)
{
  /* Force index use in case of delete/update statements
     to retrieve the latest revision of the row.
  */
  THD *thd= current_thd;
  if ((SQLCOM_UPDATE == thd->lex->sql_command)
      || (SQLCOM_DELETE == thd->lex->sql_command))
    return 0;
  return uhandler->read_time(index, ranges, rows);
}


double single_table_handler::scan_time()
{
  /* Same as read_time()
  */
  THD *thd= current_thd;
  if ((SQLCOM_UPDATE == thd->lex->sql_command)
      || (SQLCOM_DELETE == thd->lex->sql_command))
    return 0;
  return uhandler->scan_time();
}


THR_LOCK_DATA **single_table_handler::store_lock(THD *thd,
                                                 THR_LOCK_DATA **to,
                                                 enum thr_lock_type lock_type)
{
  return uhandler->store_lock(thd, to, lock_type);
}


int single_table_handler::external_lock(THD *thd, int lock_type)
{
  DBUG_ENTER("single_table_handler::external_lock");
  DBUG_RETURN(uhandler->ha_external_lock(thd, lock_type));
}


bool single_table_handler::get_error_message(int error, String *buf)
{
  switch (error) {
  case HA_ERR_REVISION_READ_ONLY_FIELD:
    buf->append(STRING_WITH_LEN("Revision field is readonly"));
    break;
  case HA_ERR_REVISION_DUPLICATE_FIELD:
    buf->append(STRING_WITH_LEN("Revision fields already set"));
    break;
  case HA_ERR_REVISION_DELETED_ROW:
    buf->append(STRING_WITH_LEN("Deleted row"));
    break;
  case HA_ERR_REVISION_MISSING_FIELD:
    buf->append(STRING_WITH_LEN("Missing revision field(s)"));
    break;
  default: { }
  }
  return TRUE;
}


int single_table_handler::alloc_records()
{
  if (!(my_multi_malloc(MYF(MY_WME),
                        &record[0], table->s->reclength,
                        &record[1], table->s->reclength,
                        NullS)))
    return (-1);
  return (0);
}


/** @brief Finds and retrieves in record[0] the latest matching revision
    given the current index if set or the first unique index otherwise.

    Key comparison is performed without the revision_id field.

    Should only be called if the table has a unique index.

 */
int single_table_handler::get_latest_revision()
{
  DBUG_ENTER("single_table_handler::get_latest_revision");

  uint key= MAX_KEY;
  int i= 0;
  bool index= false;

  if (uhandler->inited == handler::NONE)
  {
    index= true;
    if (table->s->primary_key != MAX_KEY)
      key= table->s->primary_key;
    else
    {
      for (key= 0; key < table->s->keys; ++key)
      {
        if (table->key_info[key].flags & HA_NOSAME)
          break;
      }
      if (key == table->s->keys)
        DBUG_RETURN(HA_ERR_WRONG_INDEX);
    }

    i= uhandler->ha_index_init(key, FALSE);
    if (i != 0)
      DBUG_RETURN(i);
  }
  else if (uhandler->inited == handler::INDEX)
    key= uhandler->get_index();
  
  /* Copy key_info to avoid comparing revision field
   */
  KEY key_info= table->key_info[key];
  --key_info.key_parts;
  
  /* Find last matching record.
   */
  key_part_map keypart_map=
    make_prev_keypart_map(key_info.key_parts);

  const uint key_len= key_info.key_length
    - key_info.key_part[key_info.key_parts - 1].length;

  key_copy(record[1], table->record[0], &key_info, key_len);

  i= uhandler->index_read_map(record[0], record[1], keypart_map,
                              HA_READ_PREFIX_LAST);
  if (index)
    uhandler->ha_index_end();
  DBUG_RETURN(i);
}


void single_table_handler::update_fields(uchar *record)
{
  bitmap_set_bit(table->write_set, revision_field->field_index);
  my_bitmap_map *old_map= tmp_use_all_columns(table, table->read_set);

  /* Increments the revision field in new data.
  */
  longlong old= 0;
  if (record != NULL)
  {
    my_ptrdiff_t offset= revision_field->ptr - table->record[0];
    uchar *ptr= record + offset;
    old= revision_field->val_int(ptr);
  }
  else
  {
    old= revision_field->val_int();
  }
  
  revision_field->store(old + 1);
  tmp_restore_column_map(table->read_set, old_map);

  /* Sets the timestamp field.
  */
  bitmap_set_bit(table->write_set, revision_ts_field->field_index);
  revision_ts_field->set_time();
}


void single_table_handler::copy_fields(Field **fields, const uchar *r,
                                       my_ptrdiff_t offset)
{
  my_bitmap_map *old_map= tmp_use_all_columns(table, table->write_set);
  memcpy(table->record[0], r, table->s->null_bytes);

  Field **f= fields;
  while (*f != NULL)
  {
    Field **tf= table->field;
    while ((*tf != NULL)
           && (strcmp((*tf)->field_name, (*f)->field_name) != 0))
    {
      ++tf;
    }
    if (*tf != NULL)
    {
      memcpy((*tf)->ptr, (*f)->ptr + offset, (*f)->pack_length());
    }
    ++f;
  }
  revision_del_field->store(0);
  tmp_restore_column_map(table->write_set, old_map);
}


int single_table_handler::delete_row(Field **fields, const uchar *r)
{

  copy_fields(fields, r, 0);

  /* Look for previous matching row
   */

  bool found_matching_row= FALSE;
  int i= 0;
  my_bitmap_map *old_map= tmp_use_all_columns(table, table->read_set);
  {
    int i= get_latest_revision();
    if (i == 0) {
      found_matching_row= TRUE;
    }
  }
  tmp_restore_column_map(table->read_set, old_map);

  if (!found_matching_row)
  {
    /* row not found: set deleted flag & write row
    */
    update_fields(table->record[0]);
    my_bitmap_map *old_map= tmp_use_all_columns(table, table->write_set);
    revision_del_field->store(deleted_flag);
    revision_field->store(1);

    i= write_row(table->record[0]);

    tmp_restore_column_map(table->write_set, old_map);

  }
  else
  {

    /* row is found: get revision param from previous row &
       update row with deleted flag
    */
    memcpy(table->record[1], record[0], table->s->reclength);

    bitmap_set_bit(table->write_set, revision_del_field->field_index);
    revision_del_field->store(deleted_flag);
    bitmap_clear_bit(table->write_set, revision_del_field->field_index);

    /* set revision field from previous row
       inc of revision field is done in update_row()
       date is set in update_row()
    */
    bitmap_set_bit(table->write_set, revision_field->field_index);
    my_bitmap_map *old_map= tmp_use_all_columns(table, table->read_set);
    my_ptrdiff_t offset= revision_field->ptr - table->record[0];
    uchar *ptr= table->record[1] + offset;
    revision_field->store(revision_field->val_int(ptr));
    bitmap_clear_bit(table->write_set, revision_field->field_index);

    tmp_restore_column_map(table->read_set, old_map);

    old_map= tmp_use_all_columns(table, table->write_set);
    no_field_check= TRUE;
    i= update_row(table->record[1], table->record[0]);
    no_field_check= FALSE;
    tmp_restore_column_map(table->write_set, old_map);

  }
  return i;
}


int single_table_handler::write_or_update_row(Field **fields, const uchar *old,
                                              const uchar *new_data)
{

  copy_fields(fields, old, old - new_data);

  bitmap_set_bit(table->write_set, revision_field->field_index);
  revision_field->store(1);
  bitmap_set_bit(table->write_set, revision_ts_field->field_index);
  revision_ts_field->set_time();

  uhandler->extra(HA_EXTRA_PREPARE_FOR_UPDATE);

  int r= write_row(table->record[0]);
  uhandler->extra(HA_EXTRA_REINIT_CACHE);
  if (r == HA_ERR_FOUND_DUPP_KEY)
  {

    /* Find latest revision & update revision field
    */
    int i= get_latest_revision();
    if (i != 0)
      return (i);
    {
      my_bitmap_map *old_map= tmp_use_all_columns(table, table->read_set);
      my_ptrdiff_t offset= revision_field->ptr - table->record[0];
      uchar *ptr= record[0] + offset;
      revision_field->store(revision_field->val_int(ptr));
      tmp_restore_column_map(table->read_set, old_map);
    }

    bitmap_clear_bit(table->write_set, revision_field->field_index);
    bitmap_clear_bit(table->write_set, revision_ts_field->field_index);

    my_bitmap_map *old_map= tmp_use_all_columns(table, table->write_set);
    no_field_check= TRUE;

    r= update_row(table->record[0], table->record[0]);

    no_field_check= FALSE;
    tmp_restore_column_map(table->write_set, old_map);

  }
  return r;
}


int single_table_handler::write_row(uchar * buf)
{
  DBUG_ENTER("single_table_handler::write_row");

  /* Do not check for revision fields (dump)
  */

  /* Sets the timestamp field, if not already done.
  */
  bool ts_already_set= TRUE;
  if (!bitmap_is_set(table->write_set, revision_ts_field->field_index))
  {
    bitmap_set_bit(table->write_set, revision_ts_field->field_index);
    revision_ts_field->set_time();
    ts_already_set= FALSE;
  }

  int r= uhandler->ha_write_row(buf);

  if (r == HA_ERR_FOUND_DUPP_KEY)
  {

    /* Handle the case of deleted rows.
    */

    /* inited may be NONE, but we necessarily have a unique index (dup key)
    */
    int i= get_latest_revision();
    if (i != 0)
      DBUG_RETURN(i);

    /* If deleted, increment the latest revision
    */
    if (row_is_deleted(table, record[0], revision_del_field))
    {
      my_ptrdiff_t offset= revision_field->ptr - buf;
      Field_long previous_revision
        (record[0] + offset,
         revision_field->field_length,
         revision_field->null_ptr,
         revision_field->null_bit,
         revision_field->unireg_check,
         revision_field->field_name,
         static_cast<Field_long *>(revision_field)->zerofill,
         static_cast<Field_long *>(revision_field)->unsigned_flag);
      previous_revision.table= revision_field->table;

      bitmap_set_bit(table->write_set, revision_field->field_index);
      my_bitmap_map *old_map= tmp_use_all_columns(table, table->read_set);
      revision_field->store(previous_revision.val_int() + 1);
      tmp_restore_column_map(table->read_set, old_map);
      r= uhandler->ha_write_row(buf);
    }
  }

  if (!ts_already_set)
    bitmap_clear_bit(table->write_set, revision_ts_field->field_index);
  DBUG_RETURN(r);
}


int single_table_handler::update_row(const uchar *,
                                     uchar * new_data)
{
  DBUG_ENTER("single_table_handler::update_row");

  if (compare_record(table) == FALSE)
    DBUG_RETURN(0);

  if (!no_field_check && check_revision_field())
    DBUG_RETURN(HA_ERR_REVISION_READ_ONLY_FIELD);

  int r= 0;

  /* If previous revision is deleted, an insert should be performed
     instead.
  */

  if ((uhandler->inited != handler::RND)
      && (MAX_KEY != uhandler->get_index())
      && (table->key_info[uhandler->get_index()].flags & HA_NOSAME))
  {
    r= get_latest_revision();
    if (r == HA_ERR_OUT_OF_MEM)
      DBUG_RETURN(r);
    if ((r != HA_ERR_END_OF_FILE)
        && (!no_field_check)
        && (row_is_deleted(table, record[0], revision_del_field)))
      DBUG_RETURN(HA_ERR_REVISION_DELETED_ROW);

      /* Use latest matching row to increment the revision id.
       */
      update_fields(record[0]);
  }
  else
  {
    my_bitmap_map *old_map= tmp_use_all_columns(table, table->read_set);
    const uint d= revision_del_field->val_int();
    tmp_restore_column_map(table->read_set, old_map);
    if ((deleted_flag == d)
        && (!no_field_check))
      DBUG_RETURN(HA_ERR_REVISION_DELETED_ROW);

    /* No index
    */
    update_fields();
  }

  r= uhandler->ha_write_row(new_data);

  if (MAX_KEY != uhandler->get_index())
  {
    /* Go forward once (because of newly added row).
    */
    uhandler->index_next(record[0]);
  }

  if (!no_field_check)
  {
    bitmap_clear_bit(table->write_set, revision_field->field_index);
    bitmap_clear_bit(table->write_set, revision_ts_field->field_index);
  }

  DBUG_RETURN(r);
}


int single_table_handler::delete_row(const uchar *)
{
  DBUG_ENTER("single_table_handler::delete_row");

  /* Only latest revision can be deleted.
  */
  if (check_revision_field())
    DBUG_RETURN(HA_ERR_REVISION_READ_ONLY_FIELD);

  /* Find the latest revision
     & sets the deleted flag.
   */

  int i= 0;

  /* If the table does not have a unique key, we update the current row
     without looking for subsequent revisions.
     Otherwise, we find the latest matching revision.
  */
  if ((uhandler->inited != handler::RND)
      && (MAX_KEY != uhandler->get_index())
      && (table->key_info[uhandler->get_index()].flags & HA_NOSAME))
  {
    i= get_latest_revision();
    if (i == HA_ERR_OUT_OF_MEM)
      DBUG_RETURN(i);
    if ((i != HA_ERR_END_OF_FILE)
        && (row_is_deleted(table, record[0], revision_del_field)))
      DBUG_RETURN(HA_ERR_REVISION_DELETED_ROW);

    /* Copy latest matching record in order to update the row
    */
    memcpy(table->record[0], record[0], table->s->reclength);
  }

  /* And copy the current record to perform the update.
  */
  memcpy(table->record[1], table->record[0], table->s->reclength);

  /* Check that record was not already deleted
  */
  my_bitmap_map *old_map= tmp_use_all_columns(table, table->read_set);
  if (deleted_flag == revision_del_field->val_int())
  {
    tmp_restore_column_map(table->read_set, old_map);
    DBUG_RETURN(HA_ERR_REVISION_DELETED_ROW);
  }

  bitmap_set_bit(table->write_set, revision_del_field->field_index);
  revision_del_field->store(deleted_flag);
  tmp_restore_column_map(table->read_set, old_map);

  bitmap_set_bit(table->write_set, revision_ts_field->field_index);
  revision_ts_field->set_time();

  uhandler->extra(HA_EXTRA_PREPARE_FOR_UPDATE);

  int r= uhandler->ha_update_row(table->record[1], table->record[0]);

  uhandler->extra(HA_EXTRA_REINIT_CACHE);

  bitmap_clear_bit(table->write_set, revision_field->field_index);
  bitmap_clear_bit(table->write_set, revision_ts_field->field_index);
  bitmap_clear_bit(table->write_set, revision_del_field->field_index);

  DBUG_RETURN(r);
}


void single_table_handler::delete_table(const char *name, TABLE_SHARE *table)
{
  create(table);
  uhandler->ha_delete_table(name);
}


int single_table_handler::create(const char *name, TABLE *table_arg,
                                 HA_CREATE_INFO *create_info)
{

  DBUG_ENTER("single_table_handler::create");  
  /* Comment string contains the underlying storage engine.
  */
  if (uhandler == NULL)
  {
    create(table_arg->s);
  }
  uhandler->init();

  char frm_name[FN_REFLEN];
  strxmov(frm_name, table_arg->s->path.str, reg_ext, NullS);

  bool fields_set= FALSE;
  Field* revision_id_field= NULL;

  /* If revision fields are already set, do not add them (mysqldump).
   */
  uint field_count= 0;

  /* Check that revision fields are not already there.
  */
  {
    Field **field= table_arg->s->field;
    while (*field != NULL)
    {
      bool found= FALSE;
      if ((found=(strcmp((*field)->field_name, revision_id) == 0))
          || (strcmp((*field)->field_name, revision_ts) == 0)
          || (strcmp((*field)->field_name, revision_del) == 0))
      {
        ++field_count;
        fields_set= TRUE;
        if (found)
          revision_id_field= *field;
      }
      ++field;
    }
  }

  /* If only some of them are there, it's an error
  */
  if ((field_count > 0)
      && (field_count < revision_fields_count))
     DBUG_RETURN(HA_ERR_REVISION_MISSING_FIELD);

  List<Create_field>& create_fields(current_thd->lex->alter_info.create_list);

  /* Checks that there is not already a field named revision
  */
  {
    field_count= 0;
    list_node *node= create_fields.first_node();
    while (node != create_fields.last_node())
    {
      Create_field *f= static_cast<Create_field*>(node->info);
      if ((strcmp(f->field_name, revision_id) == 0)
          || (strcmp(f->field_name, revision_ts) == 0)
          || (strcmp(f->field_name, revision_del) == 0))
      {
        fields_set= TRUE;
        ++field_count;
      }
      node= node->next;
    }
  }

  /* If only some of them are there, it's an error
  */
  if ((field_count > 0)
      && (field_count < revision_fields_count))
     DBUG_RETURN(HA_ERR_REVISION_MISSING_FIELD);

  LEX_STRING comment= { NULL, 0 };

  List<Create_field> new_create_fields;

  Create_field revision_id_create_field;

  /* Build the create_field list based on existing fields.
  */
  if (table_arg->s->fields != 0)
  {
    Create_field *prev= NULL;
    my_bitmap_map *old_map=
      tmp_use_all_columns(table_arg, table_arg->read_set);
    Field **field= table_arg->s->field;
    while (*field != NULL)
    {
      (*field)->table= table_arg;
      /* Fields are already pointing to the default values
         but ::create_field() uses record[0] (field.cc:9363)
      */
      my_ptrdiff_t diff= 
        table_arg->record[0] - table_arg->s->default_values;
      (*field)->move_field_offset(diff);
      Create_field *cf= new Create_field(*field, *field);
      (*field)->move_field_offset(-diff);

      int ts= 0;
      int ts_niladic= 0;
      uint blob= 0;
      prepare_create_field(cf, &blob, &ts, &ts_niladic,
                           uhandler->ha_table_flags());
      if (prev != NULL)
        cf->offset= prev->offset + prev->pack_length;
      else
        cf->offset= 0;
      prev= cf;
      new_create_fields.push_back(cf);
      /* Store revision id field for keys
      */
      if (strcmp((*field)->field_name, revision_id) == 0)
        revision_id_create_field= *cf;
      ++field;
    }
    tmp_restore_column_map(table_arg->read_set, old_map);
  }

  Create_field *previous= NULL;
  if (revision_id_field == NULL)
  {
    /* Adds revision columns:
       - revision_id
       - revision_ts (timestamp)
       - revision_deleted
    */

    list_node *node= new_create_fields.first_node();
    while (node->next != new_create_fields.last_node())
    {
      node= node->next;
    }

    previous= static_cast<Create_field*>(node->info);

    /* revision_id
     */
    Item_uint *default_value= new Item_uint(1);
    revision_id_create_field.init(current_thd, const_cast<char*>(revision_id),
                                  MYSQL_TYPE_LONG, NULL,
                                  NULL, NOT_NULL_FLAG | UNSIGNED_FLAG,
                                  default_value,
                                  NULL, &comment, NULL,
                                  NULL, create_info->default_table_charset,
                                  0);
    
    revision_id_create_field.key_length=
      revision_id_create_field.pack_length=
      calc_pack_length(MYSQL_TYPE_LONG, revision_id_create_field.length);
    prepare_create_field(&revision_id_create_field, NULL, NULL, NULL,
                         uhandler->ha_table_flags());

    previous= static_cast<Create_field*>(node->info);
    revision_id_create_field.offset= previous->offset + previous->pack_length;

    new_create_fields.push_back(&revision_id_create_field);
    previous= &revision_id_create_field;

    /* Update table share
     */
    table_arg->s->reclength+= revision_id_create_field.pack_length;
    table_arg->s->fields+= 1;
  }

  uint data_offset= (create_info->null_bits + 7) / 8;

  /* Create a revision_id field if it does not exist.
  */
  if (revision_id_field == NULL)
  {
    revision_id_field=
      new Field_long(table_arg->record[0] + revision_id_create_field.offset
                     + data_offset,
                     revision_id_create_field.length,
                     NULL, 0, Field::NONE, revision_id, FALSE, TRUE);

    Field **fields=
      reinterpret_cast<Field**> (alloc_root(&table_arg->s->mem_root,
                                            sizeof(Field *)
                                            * (table_arg->s->fields
                                               + revision_fields_count)));
    if (!fields)
      DBUG_RETURN(HA_ERR_OUT_OF_MEM);
    memcpy(fields, table_arg->field,
           sizeof(Field *) * (table_arg->s->fields - 1));
    fields[table_arg->s->fields + revision_fields_count - 1]= NULL;
    fields[table_arg->s->fields - 1]= revision_id_field;
    table_arg->s->field= fields;
    table_arg->field= fields;
    revision_id_field->flags|= PART_KEY_FLAG | MULTIPLE_KEY_FLAG;
    revision_id_field->field_index= table_arg->s->fields - 1;
  }

  uint i= 0;
  /* If key consists only of revision_id (case of column drop),
     remove the key.
  */
  while (i < table_arg->s->keys)
  {
    KEY *key_info= &table_arg->key_info[i];
    if ((key_info->key_parts == 1)
        && (key_info->key_part[i].fieldnr
            == revision_id_field->field_index + 1))
    {
      --table_arg->s->keys;
      if (table_arg->s->keys > i)
        memmove(&table_arg->key_info[i], &table_arg->key_info[i + 1],
                sizeof(KEY) * (table_arg->s->keys - i));
    }
    else
      ++i;
  }
  
  /* Add revision field to keys.
  */
  for (i= 0; i < table_arg->s->keys; i++)
  {

    KEY *key_info= &table_arg->key_info[i];

    /* Check is revision_id is not already set in keys
    */
    uint j;
      
    for (j= 0; j < key_info->key_parts; ++j)
    {
      if (strcmp(key_info->key_part[j].field->field_name, revision_id) == 0)
        break;
    }
    if (j < key_info->key_parts)
      continue;

    key_info->key_length+= revision_id_create_field.pack_length;
    uint old= key_info->key_parts;
    key_info->key_parts++;
    key_info->usable_key_parts++;

    /* Old key part within key_info */
    KEY_PART_INFO *key_part=
      reinterpret_cast<KEY_PART_INFO*> (alloc_root(&table_arg->s->mem_root,
                                                   sizeof(KEY_PART_INFO)
                                                   * key_info->key_parts));
    if (!key_part)
      DBUG_RETURN(HA_ERR_OUT_OF_MEM);
    memcpy(key_part, key_info->key_part,
           sizeof(KEY_PART_INFO) * old);

    KEY_PART_INFO &kpi= key_part[old];

    if (key_info->flags & HA_NULL_PART_KEY)
      revision_id_field->flags|= UNIQUE_KEY_FLAG;
    else
      revision_id_field->flags|= PRI_KEY_FLAG;

    revision_id_field->part_of_key.set_bit(i);
    revision_id_field->part_of_key_not_clustered.set_bit(i);
    revision_id_field->part_of_sortkey.set_bit(i);

    kpi.field= revision_id_field;
    kpi.offset= revision_id_create_field.offset + data_offset;
    kpi.null_offset= 0;
    kpi.length= revision_id_create_field.pack_length;
    kpi.store_length= revision_id_create_field.pack_length;
    kpi.key_type= revision_id_create_field.pack_flag;

    kpi.fieldnr= revision_id_field->field_index + 1;

    kpi.key_part_flag= 0;
    kpi.type= revision_id_field->key_type();
    kpi.null_bit= 0;

    key_info->key_part= key_part;
  }

  if (!fields_set)
  {
    /* revision_timestamp
    */
    Create_field f_ts;
    Item_int *default_value_ts= new Item_int(0);
    f_ts.init(current_thd, const_cast<char*>(revision_ts),
              MYSQL_TYPE_TIMESTAMP, NULL,
              NULL, NOT_NULL_FLAG, default_value_ts,
              NULL, &comment, NULL,
              NULL, create_info->default_table_charset,
              0);
    
    f_ts.key_length= f_ts.pack_length= calc_pack_length(MYSQL_TYPE_TIMESTAMP,
                                                        f_ts.length);
    int ts= 0;
    prepare_create_field(&f_ts, NULL, &ts, NULL, uhandler->ha_table_flags());
    f_ts.offset= previous->offset + previous->pack_length;
    previous= &f_ts;

    new_create_fields.push_back(&f_ts);

    /* Update table share
    */
    table_arg->s->reclength+= f_ts.pack_length;

    Field *revision_ts_field=
      new Field_timestamp(table_arg->record[0] + f_ts.offset + data_offset,
                          f_ts.length,
                          NULL, 0, Field::NONE, revision_ts, table_arg->s,
                          create_info->default_table_charset);

    table_arg->s->field[table_arg->s->fields]= revision_ts_field;
    table_arg->s->fields+= 1;

    /* revision_deleted flag
    */
    Create_field f_del;
    Item_uint *default_value_del= new Item_uint(0);
    f_del.init(current_thd, const_cast<char*>(revision_del),
               MYSQL_TYPE_TINY, NULL,
               NULL, NOT_NULL_FLAG | UNSIGNED_FLAG, default_value_del,
               NULL, &comment, NULL,
               NULL, create_info->default_table_charset,
               0);
    
    f_del.key_length= f_del.pack_length= calc_pack_length(MYSQL_TYPE_TINY,
                                                          f_del.length);
    prepare_create_field(&f_del, NULL, NULL, NULL, 0);
    f_del.offset= previous->offset + previous->pack_length;

    new_create_fields.push_back(&f_del);

    /* Update table share
    */
    Field *revision_del_field=
      new Field_tiny(table_arg->record[0] + f_del.offset + data_offset,
                     f_del.length, NULL, 0, Field::NONE,
                     revision_del, FALSE, TRUE);

    table_arg->s->field[table_arg->s->fields]= revision_del_field;    
    table_arg->s->reclength+= f_del.pack_length;
    table_arg->s->fields+= 1;

  }

  /* cf. my_create_frm, unireg.cc:162
     key_info_length= pack_keys(keybuff, keys, key_info, data_offset);
     with data_offset= (create_info->null_bits + 7) / 8;
 
     int2store(pos,key_part->fieldnr+1+FIELD_NAME_USED); // 483
     offset= (uint) (key_part->offset+data_offset+1); // 484
     & in open_binary_frm, table.cc:407
 
     key_part->fieldnr=	(uint16) (uint2korr(strpos) & FIELD_NR_MASK); // 546
     key_part->offset= (uint) uint2korr(strpos+2)-1; // 547

     idem for key_length:
     in pack_keys():
     int2store(pos+2,key->key_length); //unireg.cc:466

     in open_binary_frm():
     keyinfo->key_length+= HA_KEY_NULL_LENGTH; // table.cc:1073

  */

  /* Adjust keys */
  for (i= 0; i < table_arg->s->keys; ++i)
  {

    KEY *key_info= &table_arg->key_info[i];
 
    key_info->key_length-= key_info->extra_length;

    uint j;
    for (j= 0; j < key_info->key_parts; ++j)
    {
      KEY_PART_INFO &kpi= key_info->key_part[j];

      kpi.offset-= data_offset;
      kpi.fieldnr-= 1;
    }
  }

  mysql_create_frm(current_thd, frm_name,
                   table_arg->s->db.str, name, create_info,
                   new_create_fields,
                   table_arg->s->keys, table_arg->key_info,
                   uhandler);

  /* Adjust keys */
  for (i= 0; i < table_arg->s->keys; ++i)
  {

    KEY *key_info= &table_arg->key_info[i];
 
    key_info->key_length+= key_info->extra_length;

    uint j;
    for (j= 0; j < key_info->key_parts; ++j)
    {
      KEY_PART_INFO &kpi= key_info->key_part[j];

      kpi.offset+= data_offset;
      kpi.fieldnr+= 1;
    }
  }

  DBUG_RETURN(uhandler->ha_create(name, table_arg, create_info));

}
